@@ -2,6 +2,7 @@ use bumpalo::Bump;
 use rustc_hash::FxHashMap;
 
 use super::collections::*;
+use super::infinite_reevaluation_protection::InfiniteReevaluationProtector;
 use super::print_items::*;
 use super::thread_state;
 use super::writer::*;
@@ -65,17 +66,17 @@ pub struct Printer<'a> {
   writer: Writer<'a>,
   // Use a regular hash map here because only some conditions are stored (not all).
   resolved_conditions: FxHashMap<u32, Option<bool>>,
-  // Use a VecU32Map for resolved infos because it has much faster
+  // Use these "VecU32Map" for resolved infos because it has much faster
   // lookups than a hash map and generally infos seem to be resolved
   // about 90% of the time, so the extra memory usage is probably not
   // a big deal.
-  resolved_line_number_anchors: VecU32Map<u32>,
-  resolved_line_numbers: VecU32Map<u32>,
-  resolved_column_numbers: VecU32Map<u32>,
-  resolved_is_start_of_lines: VecU32Map<bool>,
-  resolved_indent_levels: VecU32Map<u8>,
-  resolved_line_start_column_numbers: VecU32Map<u32>,
-  resolved_line_start_indent_levels: VecU32Map<u8>,
+  resolved_line_number_anchors: VecU32U32Map,
+  resolved_line_numbers: VecU32U32Map,
+  resolved_column_numbers: VecU32U32Map,
+  resolved_is_start_of_lines: VecU32BoolMap,
+  resolved_indent_levels: VecU32U8Map,
+  resolved_line_start_column_numbers: VecU32U32Map,
+  resolved_line_start_indent_levels: VecU32U8Map,
   look_ahead_condition_save_points: FxHashMap<u32, &'a SavePoint<'a>>,
   look_ahead_line_number_save_points: FxHashMap<u32, &'a SavePoint<'a>>,
   look_ahead_column_number_save_points: FxHashMap<u32, &'a SavePoint<'a>>,
@@ -83,6 +84,7 @@ pub struct Printer<'a> {
   look_ahead_indent_level_save_points: FxHashMap<u32, &'a SavePoint<'a>>,
   look_ahead_line_start_column_number_save_points: FxHashMap<u32, &'a SavePoint<'a>>,
   look_ahead_line_start_indent_level_save_points: FxHashMap<u32, &'a SavePoint<'a>>,
+  infinite_reevaluation_protector: InfiniteReevaluationProtector,
   next_node_stack: RcStack,
   stored_condition_save_points: FxHashMap<u32, (&'a Condition, &'a SavePoint<'a>)>,
   max_width: u32,
@@ -111,13 +113,13 @@ impl<'a> Printer<'a> {
         },
       ),
       resolved_conditions: FxHashMap::default(),
-      resolved_line_number_anchors: VecU32Map::new(thread_state::next_line_number_anchor_id()),
-      resolved_line_numbers: VecU32Map::new(thread_state::next_line_number_id()),
-      resolved_column_numbers: VecU32Map::new(thread_state::next_column_number_id()),
-      resolved_is_start_of_lines: VecU32Map::new(thread_state::next_is_start_of_line_id()),
-      resolved_indent_levels: VecU32Map::new(thread_state::next_indent_level_id()),
-      resolved_line_start_column_numbers: VecU32Map::new(thread_state::next_line_start_column_number_id()),
-      resolved_line_start_indent_levels: VecU32Map::new(thread_state::next_line_start_indent_level_id()),
+      resolved_line_number_anchors: VecU32U32Map::with_capacity(thread_state::next_line_number_anchor_id()),
+      resolved_line_numbers: VecU32U32Map::with_capacity(thread_state::next_line_number_id()),
+      resolved_column_numbers: VecU32U32Map::with_capacity(thread_state::next_column_number_id()),
+      resolved_is_start_of_lines: VecU32BoolMap::with_capacity(thread_state::next_is_start_of_line_id()),
+      resolved_indent_levels: VecU32U8Map::with_capacity(thread_state::next_indent_level_id()),
+      resolved_line_start_column_numbers: VecU32U32Map::with_capacity(thread_state::next_line_start_column_number_id()),
+      resolved_line_start_indent_levels: VecU32U8Map::with_capacity(thread_state::next_line_start_indent_level_id()),
       look_ahead_condition_save_points: FxHashMap::default(),
       look_ahead_line_number_save_points: FxHashMap::default(),
       look_ahead_column_number_save_points: FxHashMap::default(),
@@ -125,6 +127,7 @@ impl<'a> Printer<'a> {
       look_ahead_indent_level_save_points: FxHashMap::default(),
       look_ahead_line_start_column_number_save_points: FxHashMap::default(),
       look_ahead_line_start_indent_level_save_points: FxHashMap::default(),
+      infinite_reevaluation_protector: InfiniteReevaluationProtector::with_capacity(thread_state::next_condition_reevaluation_id()),
       stored_condition_save_points: FxHashMap::default(),
       next_node_stack: RcStack::default(),
       max_width: options.max_width,
@@ -204,7 +207,7 @@ impl<'a> Printer<'a> {
       self.look_ahead_line_number_save_points.insert(line_number.unique_id(), save_point);
     }
 
-    resolved_number.copied()
+    resolved_number
   }
 
   pub fn resolved_column_number(&mut self, column_number: ColumnNumber) -> Option<u32> {
@@ -214,7 +217,7 @@ impl<'a> Printer<'a> {
       self.look_ahead_column_number_save_points.insert(column_number.unique_id(), save_point);
     }
 
-    resolved_number.copied()
+    resolved_number
   }
 
   pub fn resolved_is_start_of_line(&mut self, is_start_of_line: IsStartOfLine) -> Option<bool> {
@@ -224,7 +227,7 @@ impl<'a> Printer<'a> {
       self.look_ahead_is_start_of_line_save_points.insert(is_start_of_line.unique_id(), save_point);
     }
 
-    resolved_is_start_of_line.copied()
+    resolved_is_start_of_line
   }
 
   pub fn resolved_indent_level(&mut self, indent_level: IndentLevel) -> Option<u8> {
@@ -234,7 +237,7 @@ impl<'a> Printer<'a> {
       self.look_ahead_indent_level_save_points.insert(indent_level.unique_id(), save_point);
     }
 
-    resolved_indent_level.copied()
+    resolved_indent_level
   }
 
   pub fn resolved_line_start_column_number(&mut self, line_start_column_number: LineStartColumnNumber) -> Option<u32> {
@@ -250,7 +253,7 @@ impl<'a> Printer<'a> {
         .insert(line_start_column_number.unique_id(), save_point);
     }
 
-    resolved_line_start_column_number.copied()
+    resolved_line_start_column_number
   }
 
   pub fn resolved_line_start_indent_level(&mut self, line_start_indent_level: LineStartIndentLevel) -> Option<u8> {
@@ -266,7 +269,7 @@ impl<'a> Printer<'a> {
         .insert(line_start_indent_level.unique_id(), save_point);
     }
 
-    resolved_line_start_indent_level.copied()
+    resolved_line_start_indent_level
   }
 
   pub fn clear_info(&mut self, info: Info) {
@@ -434,17 +437,18 @@ impl<'a> Printer<'a> {
     match anchor {
       Anchor::LineNumber(anchor) => {
         let id = anchor.unique_id();
+        let current_line_number = self.writer.line_number();
         if let Some(past_line_number) = self.resolved_line_number_anchors.get(id) {
-          let current_line_number = self.writer.line_number();
-          let difference = (current_line_number as isize) - (*past_line_number as isize);
+          let difference = (current_line_number as isize) - (past_line_number as isize);
           if difference != 0 {
             let line_number_id = anchor.line_number_id();
             if let Some(value) = self.resolved_line_numbers.get(line_number_id) {
-              let new_value = ((*value as isize) + difference) as u32;
+              let new_value = ((value as isize) + difference) as u32;
               self.resolved_line_numbers.insert(line_number_id, new_value);
             }
           }
         }
+        self.resolved_line_number_anchors.insert(id, current_line_number);
       }
     }
   }
@@ -511,17 +515,27 @@ impl<'a> Printer<'a> {
   fn handle_condition_reevaluation(&mut self, condition_reevaluation: &ConditionReevaluation) {
     let condition_id = condition_reevaluation.condition_id;
     if let Some((condition, save_point)) = self.stored_condition_save_points.get(&condition_id).cloned() {
-      if let Some(resolved_condition_value) = self.resolved_conditions.get(&condition_id).and_then(|x| x.to_owned()) {
+      if let Some(past_condition_value) = self.resolved_conditions.get(&condition_id).and_then(|x| x.to_owned()) {
         self.resolving_save_point.replace(save_point);
         let mut context = ConditionResolverContext::new(self, save_point.writer_state.writer_info(self.writer.indent_width()));
-        let condition_value = condition.resolve(&mut context);
+        let latest_condition_value = condition.resolve(&mut context);
         self.resolving_save_point.take();
-        if let Some(condition_value) = condition_value {
-          if condition_value != resolved_condition_value {
-            self.update_state_to_save_point(save_point, false);
+
+        // Do not re-evaluate the condition if it's flipped back and forth a decent number of times.
+        // If it hits the max number of times it can flip then an error will be logged.
+        let should_reevaluate = self.infinite_reevaluation_protector.should_reevaluate(
+          condition_reevaluation.condition_reevaluation_id,
+          latest_condition_value,
+          past_condition_value,
+        );
+        if should_reevaluate {
+          if let Some(latest_condition_value) = latest_condition_value {
+            if latest_condition_value != past_condition_value {
+              self.update_state_to_save_point(save_point, false);
+            }
+          } else {
+            self.resolved_conditions.remove(&condition_id);
           }
-        } else {
-          self.resolved_conditions.remove(&condition_id);
         }
       }
     }
