@@ -1,3 +1,5 @@
+use std::rc::Rc;
+
 use dprint_core::formatting::*;
 
 enum Node<'a> {
@@ -149,16 +151,23 @@ fn gen_node(node: Node) -> PrintItems {
 
 fn gen_array_literal_expression(expr: &ArrayLiteralExpression) -> PrintItems {
   let mut items = PrintItems::new();
-  let start_info = Info::new("start");
-  let end_info = Info::new("end");
+  let start_ln = LineNumber::new("start");
+  let end_ln = LineNumber::new("end");
   let is_multiple_lines = create_is_multiple_lines_resolver(
     expr.position.clone(),
     expr.elements.iter().map(|e| e.position.clone()).collect(),
-    start_info,
-    end_info,
+    start_ln,
+    end_ln,
   );
 
-  items.push_info(start_info);
+  // actions::if_column_number_changes is a helper that uses lower level IR to tell when the column number
+  // changes at this point
+  items.extend(actions::if_column_number_changes(move |context| {
+    context.clear_info(end_ln);
+  }));
+
+  items.push_info(start_ln);
+  items.push_anchor(LineNumberAnchor::new(end_ln)); // updates the line number of end_ln when this changes
 
   items.push_str("[");
   items.push_condition(conditions::if_true("arrayStartNewLine", is_multiple_lines.clone(), Signal::NewLine.into()));
@@ -174,11 +183,11 @@ fn gen_array_literal_expression(expr: &ArrayLiteralExpression) -> PrintItems {
   items.push_condition(conditions::if_true("arrayEndNewLine", is_multiple_lines, Signal::NewLine.into()));
   items.push_str("]");
 
-  items.push_info(end_info);
+  items.push_info(end_ln);
 
   return items;
 
-  fn gen_elements(elements: &[ArrayElement], is_multiple_lines: &(impl Fn(&mut ConditionResolverContext) -> Option<bool> + Clone + 'static)) -> PrintItems {
+  fn gen_elements(elements: &[ArrayElement], is_multiple_lines: &ConditionResolver) -> PrintItems {
     let mut items = PrintItems::new();
     let elements_len = elements.len();
 
@@ -206,15 +215,8 @@ fn gen_array_element(element: &ArrayElement) -> PrintItems {
 
 // helper functions
 
-fn create_is_multiple_lines_resolver(
-  parent_position: Position,
-  child_positions: Vec<Position>,
-  start_info: Info,
-  end_info: Info,
-) -> impl Fn(&mut ConditionResolverContext) -> Option<bool> + Clone + 'static {
-  // This could be more efficient by only using references and avoid clones
-  // I'm too lazy to update this sample, but it should help you get the idea.
-  move |condition_context: &mut ConditionResolverContext| {
+fn create_is_multiple_lines_resolver(parent_position: Position, child_positions: Vec<Position>, start_ln: LineNumber, end_ln: LineNumber) -> ConditionResolver {
+  Rc::new(move |condition_context: &mut ConditionResolverContext| {
     // no items, so format on the same line
     if child_positions.is_empty() {
       return Some(false);
@@ -226,6 +228,6 @@ fn create_is_multiple_lines_resolver(
     }
 
     // check if it spans multiple lines, and if it does then make it multi-line
-    condition_resolvers::is_multiple_lines(condition_context, &start_info, &end_info)
-  }
+    condition_helpers::is_multiple_lines(condition_context, start_ln, end_ln)
+  })
 }
